/*
 * Compaq Hot Plug Controller Driver
 *
 * Copyright (c) 1995,2001 Compaq Computer Corporation
 * Copyright (c) 2001 Greg Kroah-Hartman (greg@kroah.com)
 * Copyright (c) 2001 IBM Corp.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, GOOD TITLE or
 * NON INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Send feedback to <greg@kroah.com>
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/proc_fs.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <asm/uaccess.h>
#include "hotplug_pci.h"


#if !defined(CONFIG_HOTPLUG_PCI_MODULE)
	#define MY_NAME	"hotplug_pci.o"
#else
	#define MY_NAME	THIS_MODULE->name
#endif

#define dbg(fmt, arg...) do { if (debug) printk(KERN_DEBUG "%s: "__FUNCTION__": " fmt , MY_NAME , ## arg); } while (0)
#define err(format, arg...) printk(KERN_ERR "%s: " format , MY_NAME , ## arg)
#define info(format, arg...) printk(KERN_INFO "%s: " format , MY_NAME , ## arg)
#define warn(format, arg...) printk(KERN_WARNING "%s: " format , MY_NAME , ## arg)


/* local variables */
static int debug;

#define DRIVER_VERSION	"0.1"
#define DRIVER_AUTHOR	"Dan Zink <dan.zink@compaq.com>, Greg Kroah-Hartman <greg@kroah.com>"
#define DRIVER_DESC	"Hot Plug PCI Core Driver"

#define HOTPLUG_PCI_MODULE_MINOR 208


/*
This is a mechanism to notify user level processes asynchronously of events
that happened on a driver level.  The mechanism transmits messages by
dispatching SIGIO and then expects the application to read the message in a
messages queue.
 */

static struct fasync_struct *fasync;	/* = NULL */

static struct semaphore notify_sem;	/* mutex for process loop (only up if something to process) */
static struct semaphore exit_sem;	/* guard to make sure thread has exited before calling it quits */

static struct semaphore queue_sem;	/* locks the queue to prevent new messages from being written in parallel */
static struct semaphore data_sem;	/* protects msg_buffer to prevent new messages from being read or written in parallel */

static struct hotplug_pci_message msg_buffer[HPCD_BUFFER_LENGTH];
static int msg_start_index;
static int msg_serial_number;
static int hpcd_finished;


LIST_HEAD(hotplug_pci_controller_list);


/**
 * pci_hp_register - register a hotplug_controller with the PCI hotplug controller subsystem
 * @ctrl: pointer to the &struct hotplug_controller to register
 *
 * Registers a hotplug controller with the hotplug controller
 * subsystem, which will allow userspace interaction to the
 * controller.
 *
 * Returns 0 if successful, anything else for an error.
 */
int pci_hp_register (struct hotplug_controller *ctrl)
{
	list_add (&ctrl->controller_list, &hotplug_pci_controller_list);
	info ("Hotplug PCI support registered for %s\n", ctrl->name);
	return 0;
}

/**
 * pci_hp_deregister - deregister a hotplug_controller with the PCI hotplug controller subsystem
 * @ctrl: pointer to the &struct hotplug_controller to register
 *
 * The @ctrl must have been registered with the hotplug controller
 * subsystem previously with a call to pci_hp_register().
 *
 * Returns 0 if successful, anything else for an error.
 */
int pci_hp_deregister (struct hotplug_controller *ctrl)
{
	list_del (&ctrl->controller_list);
	info ("Hotplug PCI deregistering %s\n", ctrl->name);
	return 0;
}


static int build_dev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, struct pci_dev **pci_dev)
{
	struct pci_dev *my_dev;
	struct pci_bus *my_bus;

	/* Some validity checks. */
	if ((function > 7) ||
	    (slot > 31) ||
	    (pci_dev == NULL) ||
	    (ops == NULL))
		return -ENODEV;

	my_dev = kmalloc (sizeof (struct pci_dev), GFP_KERNEL);
	if (!my_dev)
		return -ENOMEM;
	my_bus = kmalloc (sizeof (struct pci_bus), GFP_KERNEL);
	if (!my_bus) {
		kfree (my_dev);
		return -ENOMEM;
	}
	memset(my_dev, 0, sizeof(struct pci_dev));
	memset(my_bus, 0, sizeof(struct pci_bus));

	my_bus->number = bus;
	my_bus->ops = ops;
	my_dev->devfn = PCI_DEVFN(slot, function);
	my_dev->bus = my_bus;
	*pci_dev = my_dev;
	return 0;
}

/**
 * pci_read_config_byte_nodev - read a byte from a pci device
 * @ops: pointer to a &struct pci_ops that will be used to read from the pci device
 * @bus: the bus of the pci device to read from
 * @slot: the pci slot number of the pci device to read from
 * @function: the function of the pci device to read from
 * @where: the location on the pci address space to read from
 * @value: pointer to where to place the data read
 *
 * Like pci_read_config_byte() but works for pci devices that do not have a
 * pci_dev structure set up yet.
 * Returns 0 on success.
 */
int pci_read_config_byte_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u8 *value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %p\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_read_config_byte (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_read_config_byte(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/**
 * pci_read_config_word_nodev - read a word from a pci device
 * @ops: pointer to a &struct pci_ops that will be used to read from the pci device
 * @bus: the bus of the pci device to read from
 * @slot: the pci slot number of the pci device to read from
 * @function: the function of the pci device to read from
 * @where: the location on the pci address space to read from
 * @value: pointer to where to place the data read
 *
 * Like pci_read_config_word() but works for pci devices that do not have a
 * pci_dev structure set up yet. 
 * Returns 0 on success.
 */
int pci_read_config_word_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u16 *value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %p\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_read_config_word (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_read_config_word(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/**
 * pci_read_config_dword_nodev - read a dword from a pci device
 * @ops: pointer to a &struct pci_ops that will be used to read from the pci device
 * @bus: the bus of the pci device to read from
 * @slot: the pci slot number of the pci device to read from
 * @function: the function of the pci device to read from
 * @where: the location on the pci address space to read from
 * @value: pointer to where to place the data read
 *
 * Like pci_read_config_dword() but works for pci devices that do not have a
 * pci_dev structure set up yet. 
 * Returns 0 on success.
 */
int pci_read_config_dword_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u32 *value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %p\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_read_config_dword (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_read_config_dword(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/**
 * pci_write_config_byte_nodev - write a byte to a pci device
 * @ops: pointer to a &struct pci_ops that will be used to write to the pci device
 * @bus: the bus of the pci device to write to
 * @slot: the pci slot number of the pci device to write to
 * @function: the function of the pci device to write to
 * @where: the location on the pci address space to write to
 * @value: the value to write to the pci device
 *
 * Like pci_write_config_byte() but works for pci devices that do not have a
 * pci_dev structure set up yet. 
 * Returns 0 on success.
 */
int pci_write_config_byte_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u8 value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %d\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_write_config_byte (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_write_config_byte(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/**
 * pci_write_config_word_nodev - write a word to a pci device
 * @ops: pointer to a &struct pci_ops that will be used to write to the pci device
 * @bus: the bus of the pci device to write to
 * @slot: the pci slot number of the pci device to write to
 * @function: the function of the pci device to write to
 * @where: the location on the pci address space to write to
 * @value: the value to write to the pci device
 *
 * Like pci_write_config_word() but works for pci devices that do not have a
 * pci_dev structure set up yet. 
 * Returns 0 on success.
 */
int pci_write_config_word_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u16 value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %d\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_write_config_word (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_write_config_word(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/**
 * pci_write_config_dword_nodev - write a dword to a pci device
 * @ops: pointer to a &struct pci_ops that will be used to write to the pci device
 * @bus: the bus of the pci device to write to
 * @slot: the pci slot number of the pci device to write to
 * @function: the function of the pci device to write to
 * @where: the location on the pci address space to write to
 * @value: the value to write to the pci device
 *
 * Like pci_write_config_dword() but works for pci devices that do not have a
 * pci_dev structure set up yet. 
 * Returns 0 on success.
 */
int pci_write_config_dword_nodev (struct pci_ops *ops, u8 bus, u8 slot, u8 function, int where, u32 value)
{
	struct pci_dev *dev = NULL;
	int result;

	dbg("%p, %d, %d, %d, %d, %d\n", ops, bus, slot, function, where, value);
	dev = pci_find_slot(bus, PCI_DEVFN(slot, function));
	if (dev) {
		dbg("using native pci_dev\n");
		return pci_write_config_dword (dev, where, value);
	}
	
	result = build_dev (ops, bus, slot, function, &dev);
	if (result)
		return result;
	result = pci_write_config_dword(dev, where, value);
	kfree (dev->bus);
	kfree (dev);
	return result;
}

/*
 * This is code that scans the pci buses.
 * Every bus and every function is presented to a custom
 * function that can act upon it.
 */

static int pci_visit_bus(struct pci_visit * fn, struct pci_bus_wrapped *wrapped_bus, struct pci_dev_wrapped *wrapped_parent)
{
	struct list_head *ln;
	struct pci_dev *dev;
	struct pci_dev_wrapped wrapped_dev;
	int result = 0;

	dbg("scanning bus %02x\n", wrapped_bus->bus->number);

	if (fn->pre_visit_pci_bus) {
		result = fn->pre_visit_pci_bus(wrapped_bus, wrapped_parent);
		if (result)
			return result;
	}

	ln = wrapped_bus->bus->devices.next; 
	while (ln != &wrapped_bus->bus->devices) {
		dev = pci_dev_b(ln);
		ln = ln->next;

		memset(&wrapped_dev, 0, sizeof(struct pci_dev_wrapped));
		wrapped_dev.dev = dev;

		result = pci_visit_dev(fn, &wrapped_dev, wrapped_bus);
		if (result)
			return result;
	}

	if (fn->post_visit_pci_bus)
		result = fn->post_visit_pci_bus(wrapped_bus, wrapped_parent);

	return result;
}


static int pci_visit_bridge(struct pci_visit * fn, struct pci_dev_wrapped *wrapped_dev, struct pci_bus_wrapped *wrapped_parent)
{
	struct pci_bus *bus = wrapped_dev->dev->subordinate;
	struct pci_bus_wrapped wrapped_bus;
	int result;

	memset(&wrapped_bus, 0, sizeof(struct pci_bus_wrapped));
	wrapped_bus.bus = bus;

	dbg("scanning bridge %02x, %02x\n", wrapped_dev->dev->devfn >> 3, wrapped_dev->dev->devfn & 0x7);

	if (fn->visit_pci_dev) {
		result = fn->visit_pci_dev(wrapped_dev, wrapped_parent);
		if (result)
			return result;
	}

	result = pci_visit_bus(fn, &wrapped_bus, wrapped_dev);
	return result;
}


int pci_visit_dev(struct pci_visit *fn, struct pci_dev_wrapped *wrapped_dev, struct pci_bus_wrapped *wrapped_parent)
{
	struct pci_dev* dev = wrapped_dev ? wrapped_dev->dev : NULL;
	int result = 0;

	if (!dev)
		return 0;

	if (fn->pre_visit_pci_dev) {
		result = fn->pre_visit_pci_dev(wrapped_dev, wrapped_parent);
		if (result)
			return result;
	}

	switch (dev->class >> 8) {
		case PCI_CLASS_BRIDGE_PCI:
			result = pci_visit_bridge(fn, wrapped_dev, wrapped_parent);
			if (result)
				return result;
			break;
		default:
			dbg("scanning device %02x, %02x\n", PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn));
			if (fn->visit_pci_dev) {
				result = fn->visit_pci_dev(wrapped_dev, wrapped_parent);
				if (result)
					return result;
			}
	}

	if (fn->post_visit_pci_dev)
		result = fn->post_visit_pci_dev(wrapped_dev, wrapped_parent);

	return result;
}


/**
 * pci_hp_notify_id - adds a message to hotplug controller subsystem
 * @id: the message id to add
 *
 * Adds a message to the hotplug controller subsystem, which is then
 * sent to any user space process that is watching on the /dev node.
 *
 */
void pci_hp_notify_id (int id)
{
	dbg("setting up message (%d, %d)\n", msg_serial_number, id);

	down (&queue_sem);
	down (&data_sem);

	msg_buffer[msg_start_index].eb_serial_number = msg_serial_number;
	msg_buffer[msg_start_index].message_id = id;

	msg_serial_number++;
	msg_start_index = (msg_start_index + 1) % HPCD_BUFFER_LENGTH;
	
	up (&data_sem);
	up (&notify_sem);
	up (&queue_sem);
}


static int hotplug_pci_start_thread (void *data)
{
	lock_kernel();
	// release insmod memory map, use the kernel's mm
	// init is session and group leader
	// disconnect filesystem usage
	daemonize();
	// New name
	strcpy(current->comm, "phpd_notify");
	unlock_kernel();

	hpcd_finished = 0;
	while (1) {
		if (down_interruptible(&notify_sem))
			break;
		dbg("woken finished = %d\n", hpcd_finished);
		if (hpcd_finished)
			break;

		if (fasync) {
			down (&queue_sem);
			kill_fasync(&fasync, SIGIO, POLL_IN);
			up (&queue_sem);
		}
	}
	dbg("signals exit\n");
	up(&exit_sem);
	return 0;
}


static int hotplug_pci_start_notify (void)
{
	int pid;

        init_MUTEX(&queue_sem);
	init_MUTEX(&data_sem);
	init_MUTEX_LOCKED(&notify_sem);
	init_MUTEX_LOCKED(&exit_sem);
	memset(&msg_buffer, 0, sizeof(struct hotplug_pci_message) * HPCD_BUFFER_LENGTH);
	msg_start_index = 0;
	msg_serial_number = 0;
	hpcd_finished = 0;
	
	pid = kernel_thread(hotplug_pci_start_thread, 0, 0);
	if (pid < 0) {
		err ("Can't start up our notifier thread\n");
		return -1;
	}
	dbg("Our notifier thread pid = %d\n", pid);
	return 0;
}


static void hotplug_pci_stop_notify (void)
{
	dbg("thread finish command given\n");
	down (&queue_sem);
	hpcd_finished = 1;
	up (&notify_sem);
	up (&queue_sem);
	
	dbg("wait for hpcd_notifier to exit\n");
	down (&exit_sem);
	dbg("wait for hpcd_notifier exited\n");
}


static struct hotplug_controller *get_controller_from_slot (int physical_slot)
{
	struct hotplug_controller *ctrl;
	struct hotplug_controller *new_ctrl;
	struct list_head *tmp;

	list_for_each (tmp, &hotplug_pci_controller_list) {
		ctrl = list_entry (tmp, struct hotplug_controller, controller_list);
		new_ctrl = ctrl->get_controller_from_slot(ctrl, physical_slot);
		if (new_ctrl) {
			return new_ctrl;
		}
	}
	return NULL;
}


static int do_ioctl (struct hotplug_pci_ioctl_data *stuff)
{
	struct hotplug_controller *ctrl;
	int result = 0;

	dbg("stuff->type_of_request = %d\n", stuff->type_of_request);
	dbg("stuff->physical_slot =   %d\n", stuff->physical_slot);
	dbg("stuff->value =           %d\n", stuff->value);
	
	/*
	 * Since everything we do shouldn't be done by "ordinary" users,
	 * make sure we have that capability.
	 */
	if (!capable(CAP_SYS_RAWIO))
		return -EPERM;

	ctrl = get_controller_from_slot(stuff->physical_slot);
	dbg("ctrl = %p\n", ctrl);
	if (!ctrl)
		return -ENODEV;

	switch (stuff->type_of_request) {
		case PCI_HP_IOCTL_REQUEST_SET_ATTENTION_STATUS:
			result = ctrl->set_attention_status(ctrl, stuff->physical_slot, stuff->value);
			break;

		case PCI_HP_IOCTL_REQUEST_ADD_BOARD:
			result = ctrl->process_SI(ctrl, stuff->physical_slot);
			break;

		case PCI_HP_IOCTL_REQUEST_POWER_DOWN_BOARD:
			result = ctrl->process_SS(ctrl, stuff->physical_slot);
			break;

		case PCI_HP_IOCTL_REQUEST_HARDWARE_TEST:
			result = ctrl->hardware_test(ctrl, stuff->physical_slot, stuff->value);
			break;
		default:
			result = -ENOTTY;
	}
	return result;
}


static int hotplug_pci_ioctl (struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct hotplug_pci_ioctl_data *stuff;
	int result = -ENOTTY;

	dbg("(cmd =%d)\n", cmd);

	switch (cmd) {
		case GUI_BACK_DOOR:
			stuff = kmalloc (sizeof (struct hotplug_pci_ioctl_data), GFP_KERNEL);
			if (!stuff)
				return -ENOMEM;

			if (copy_from_user(stuff, (void *)arg, sizeof(struct hotplug_pci_ioctl_data))) {
				kfree (stuff);
				return -EFAULT;
			}

			result = do_ioctl(stuff);

			if (copy_to_user((void *)arg, stuff, sizeof(struct hotplug_pci_ioctl_data)))
				result = -EFAULT;

			kfree (stuff);
			break;

		default:
			break;
	}
	return result;
}


static int hotplug_pci_open (struct inode *inode, struct file *filp)
{
	dbg("initializing reader at %d\n", msg_start_index);
	filp->private_data = (void*)(msg_start_index);
	return 0;
}


static int hotplug_pci_close (struct inode *inode, struct file *filp)
{
	return 0;
}


static int hotplug_pci_fasync(int fd, struct file *filp, int mode)
{
	int result;

	dbg("entering\n");
	if (!mode) {
		down (&queue_sem);
		result = fasync_helper(fd, filp, 0, &fasync);
		up (&queue_sem);
	}
	else {
		down (&queue_sem);
		result = fasync_helper(fd, filp, 1, &fasync);
		up (&queue_sem);
	}
	
	dbg("leaving\n");
	return result;
}


static ssize_t hotplug_pci_read(struct file *filp, char *buf, size_t count, loff_t * ppos)
{
	int curr = (int)filp->private_data;

	dbg("reading queue element %d\n", curr);

	down (&data_sem);
	if (buf) {
		dbg("reading message (%d, %d)\n", msg_buffer[curr].eb_serial_number, msg_buffer[curr].message_id); 
		copy_to_user(buf, &msg_buffer[curr], sizeof(struct hotplug_pci_message));
	}
	filp->private_data = (void*) ((curr + 1) % HPCD_BUFFER_LENGTH);
	up (&data_sem);

	return sizeof(struct hotplug_pci_message);
}


static struct file_operations hotplug_pci_fops = {
	owner:		THIS_MODULE,
	read:		&hotplug_pci_read,
	ioctl:		&hotplug_pci_ioctl,
	open:		&hotplug_pci_open,
	release:	&hotplug_pci_close,
	fasync:		&hotplug_pci_fasync,
};


static struct miscdevice hotplug_pci_miscdev = {
	minor:		HOTPLUG_PCI_MODULE_MINOR,
	name:		"hotplug_pci",
	fops:		&hotplug_pci_fops,
};


static int __init hotplug_pci_init (void)
{
	int result;

	result = misc_register (&hotplug_pci_miscdev);
	if (result < 0) {
		warn("can't get minor %d\n", HOTPLUG_PCI_MODULE_MINOR);
		dbg("misc_register returned %d\n", result);
		return result;
	}

	result = hotplug_pci_start_notify();
	if (result)
		goto error;

	info (DRIVER_DESC " version: " DRIVER_VERSION "\n");

error:
	return result;
}


static void __exit hotplug_pci_exit (void)
{
	int result;

	hotplug_pci_stop_notify ();
	result = misc_deregister (&hotplug_pci_miscdev);
	if (result)
		dbg("misc_deregister returned %d\n", result);
}

module_init(hotplug_pci_init);
module_exit(hotplug_pci_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debugging mode enabled or not");


EXPORT_SYMBOL(pci_hp_notify_id);
EXPORT_SYMBOL(pci_hp_register);
EXPORT_SYMBOL(pci_hp_deregister);

EXPORT_SYMBOL(pci_visit_dev);

EXPORT_SYMBOL(pci_read_config_byte_nodev);
EXPORT_SYMBOL(pci_read_config_word_nodev);
EXPORT_SYMBOL(pci_read_config_dword_nodev);

EXPORT_SYMBOL(pci_write_config_byte_nodev);
EXPORT_SYMBOL(pci_write_config_word_nodev);
EXPORT_SYMBOL(pci_write_config_dword_nodev);

