/*
 * Compaq Hot Plug Controller Driver
 *
 * Copyright (c) 1995,2001 Compaq Computer Corporation
 * Copyright (c) 2001 Greg Kroah-Hartman (greg@kroah.com)
 * Copyright (c) 2001 IBM Corp.
 *
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, GOOD TITLE or
 * NON INFRINGEMENT.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Send feedback to <greg@kroah.com>
 *
 */
#ifndef _HOTPLUG_PCI_H
#define _HOTPLUG_PCI_H


struct hotplug_controller {
	struct list_head	controller_list;

	/*
	 * Pointer to a string describing the hotplug pci controller.  This
	 * will show up in the kernel log when the device is initialized and
	 * removed.
	 */
	const char *name;

	/**
	 * get_controller_from_slot - called to determine if this controller owns this slot
	 * 
	 * @ctrl: pointer to the controller structure
	 * @physical_slot: number of the physical slot that is being asked for.
	 *
	 * This is called by the hotplug core to determine if this controller
	 * is in charge of the @physical_slot.
	 *
	 * Return pointer to the hotplug_controller structure if it is in
	 * charge of this slot, otherwise returns NULL.
	 */
	struct hotplug_controller * (* get_controller_from_slot) (struct hotplug_controller *ctrl, int physical_slot);

	/**
	 * set_attention_status - called to set the specific slots LED
	 *
	 * @ctrl: pointer to the controller structure
	 * @physical_slot: number of the physical slot of the LED
	 * @status: status to set the LED to.
	 *
	 * Called by the hotplug pci core when a user wants to set a specific
	 * LED to a specific value.
	 *
	 * Return 0 if success, otherwise an error number.
	 */
	int (* set_attention_status) (struct hotplug_controller *ctrl,
				      u32 physical_slot, u32 status);

	/**
	 * process_SI - handle a slot enable
	 *
	 * @ctrl: pointer to the hotplug controler structure
	 * @physical_slot: number of the slot to enable
	 *
	 * Called by the hotplug pci core when a user wants to activate a
	 * specific pci slot.
	 *
	 * Return 0 if the card in the slot is successfully initialized,
	 * otherwise an error number.
	 */
	int (* process_SI) (struct hotplug_controller *ctrl, u32 physical_slot);

	/**
	 * process_SS - handle a slot disable
	 *
	 * @ctrl: pointer to the hotplug controler structure
	 * @physical_slot: number of the slot to disable
	 *
	 * Called by the hotplug pci core when a user wants to disable a
	 * specific pci slot.
	 *
	 * Return 0 if the card in the slot is successfully disabled, otherwise
	 * an error number.
	 */
	int (* process_SS) (struct hotplug_controller *ctrl, u32 physical_slot);

	/**
	 * hardware_test - implement a controller specific hardware test
	 *
	 * @ctrl: pointer to the hotplug controller structure
	 * @physical_slot: number of the slot to perform the test on.
	 * @value: the specific test to perform
	 *
	 * Called when the user asks for a hardware test to be done on a
	 * specific slot.
	 *
	 * Return 0 if the test is successful, otherwise an error number.
	 */
	int (* hardware_test) (struct hotplug_controller *ctrl, 
			       u32 physical_slot, u32 value);

	void *private;
};


struct hotplug_pci_ioctl_data {
	u32	type_of_request;
	u32	physical_slot;
	u32	value;
};


struct hotplug_pci_message {
	int	eb_serial_number;
	int	message_id;
};


/* this number needs adjusting to the maximum amount of pending messages
 * possible.
 * If there is no upper bound on this, the implementation here is rather 
 * dangerous to use, as messages are not prevented to "step over" each
 * other.
 */
#define HPCD_BUFFER_LENGTH	32

#define PCI_HP_IOCTL_REQUEST_SET_ATTENTION_STATUS	1
#define PCI_HP_IOCTL_REQUEST_ADD_BOARD			2
#define PCI_HP_IOCTL_REQUEST_POWER_DOWN_BOARD		3
#define PCI_HP_IOCTL_REQUEST_HARDWARE_TEST		4

/* Our ioctl command */
/* FIXME!  put something real here... gregkh */
#define GUI_BACK_DOOR		28


extern void pci_hp_notify_id	(int id);
extern int pci_hp_register	(struct hotplug_controller *ctrl);
extern int pci_hp_deregister	(struct hotplug_controller *ctrl);



struct pci_dev_wrapped {
	struct pci_dev* dev;
	void* data;
};

struct pci_bus_wrapped {
	struct pci_bus* bus;
	void* data;
};

struct pci_visit {
	int (* pre_visit_pci_bus)	(struct pci_bus_wrapped *, struct pci_dev_wrapped *);
	int (* post_visit_pci_bus)	(struct pci_bus_wrapped *, struct pci_dev_wrapped *);

	int (* pre_visit_pci_dev)	(struct pci_dev_wrapped *, struct pci_bus_wrapped *);
	int (* visit_pci_dev)		(struct pci_dev_wrapped *, struct pci_bus_wrapped *);
	int (* post_visit_pci_dev)	(struct pci_dev_wrapped *, struct pci_bus_wrapped *);
};

extern int pci_visit_dev	(struct pci_visit *fn, struct pci_dev_wrapped *wrapped_dev, struct pci_bus_wrapped *wrapped_parent);

extern int pci_read_config_byte_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u8 *val);
extern int pci_read_config_word_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u16 *val);
extern int pci_read_config_dword_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u32 *val);

extern int pci_write_config_byte_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u8 val);
extern int pci_write_config_word_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u16 val);
extern int pci_write_config_dword_nodev (struct pci_ops *ops, u8 bus, u8 device, u8 function, int where, u32 val);



#endif

